
import polyscope as ps
import numpy as np 
from matplotlib import cm



def separate_faces_with_quantity(Faces_q,n_towers=10): 
    n_towers = int(np.amax(Faces_q[:,[3,4]])+1)
   
    Occupancy=np.zeros(n_towers)
    Dict={}
    Faces_separated=[]
    for face in Faces_q : 
        num1,num2 = face[[3,4]].astype(int)
        if num1!=-1:
            if Occupancy[num1]==0:
                Dict[num1]=[face[[0,1,2,5]]]
                Occupancy[num1]+=1
            else : 
                Dict[num1].append(face[[0,1,2,5]])
            
        if num2!=-1:
            if Occupancy[num2]==0:
                Dict[num2]=[face[[0,1,2,5]]]
                Occupancy[num2]+=1
            else : 
                Dict[num2].append(face[[0,1,2,5]])
            
    for i in sorted(list(Dict.keys())) : 
        Faces_separated.append(np.array(Dict[i]))
        
    return(np.array(Faces_separated))



##Support functions, simple mesh


def view_faces_values_on_embryo(Mesh,Vf,name_values = "Values",ps_mesh = None,colormap = cm.jet,min_to_zero=True,clean_before = True, clean_after=True,show=True,highlight_junctions=False,adapt_values = True, scattered = False):
    
    if scattered : 
        v,f,idx = Mesh.v_scattered,Mesh.f_scattered, Mesh.idx_scattered
    
    else : 
        v,f,idx = Mesh.v,Mesh.f, np.arange(len(Mesh.f))
        
    Values = Vf.copy()
    if adapt_values : 
        if min_to_zero : 
            Values-=np.amin(Values)
        Values/=(np.amax(Values)-np.amin(Values))

    Values = Values[idx]
    colors_face = colormap(Values)[:,:3]

    ps.init()
    
    if clean_before : 
        ps.remove_all_structures()


    if ps_mesh == None : 
        ps_mesh = ps.register_surface_mesh("Embryo", v,f[:,[0,1,2]])

    ps_mesh.set_color((0.3, 0.6, 0.8)) # rgb triple on [0,1]
    #ps_mesh.set_transparency(0.2)
    ps_mesh.add_color_quantity(name_values, colors_face, defined_on='faces',enabled=True)
    
    
    if highlight_junctions : 
        plot_trijunctions(Mesh,clean_before = False, clean_after = False, show=False, color = "uniform",value_color = np.ones(3))
    
    ps.set_ground_plane_mode("none") 
    
    if show : 
        ps.show()
    
    if clean_after : 
        ps.remove_all_structures()
        
    return(ps_mesh)
        

def view_vertex_values_on_embryo(Mesh,H,remove_trijunctions=True,ps_mesh = None,name_values = "Values",scattering_coeff = 1.0,clean_before = True, clean_after=True,show=True,highlight_junction = False,scattered = True):
    #H : Gaussian curvature at each vertex

    v,f = Mesh.v,Mesh.f

    Values = np.zeros(len(Mesh.f))

    Mesh.mark_trijunctional_vertices()
    valid_values = []
    indices_nan = []
    for i,face in enumerate(f[:,[0,1,2]]) : 
        a,b,c = face
        liste = []
        for vert_idx in face : 
            if remove_trijunctions : 
                if not Mesh.vertices[vert_idx].on_trijunction :
                    liste.append(H[vert_idx])
            else : 
                liste.append(H[vert_idx])
        if len(liste)>0 : 
            Values[i]=np.mean(np.array(liste))
            valid_values.append(Values[i])
        else : 
            indices_nan.append(i)

    mean = np.mean(np.array(valid_values))
    for i in indices_nan : 
        Values[i]=mean

    ps.init()

    Values-=np.amin(Values)
    Values/=np.amax(Values)

    if scattered :
        v,f,idx = Mesh.v_scattered,Mesh.f_scattered,Mesh.idx_scattered
        Values = Values[idx]

    colors_face = cm.jet(Values)[:,:3]

    if ps_mesh == None : 
        ps_mesh = ps.register_surface_mesh("Embryo", v,f[:,[0,1,2]])
    ps_mesh.add_color_quantity(name_values, colors_face, defined_on='faces',enabled=True)
       
    if highlight_junction : 
        plot_trijunctions(Mesh,clean_before = False, clean_after = False, show=False, color = "uniform",value_color = np.ones(3))

    ps.set_ground_plane_mode("none") 
    if show : 
        ps.show()
    
    if clean_after : 
        ps.remove_all_structures()
    
    return(ps_mesh)
    
def view_dict_values_on_mesh(Mesh,dict_values,alpha = 0.05,ps_mesh = None, clean_before = True, clean_after=True,show=True,scattered = False,name_values = "Values",alpha_values=True, min_value = None, max_value = None,cmap = cm.jet):

    v,f = Mesh.v,Mesh.f

    def find_values(face): 
        return(dict_values[tuple(face[[3,4]])])
    values = np.array(list(map(find_values,f)))
    values_values = np.array(list(dict_values.values()))
    if alpha_values:
        mint = np.quantile(values_values,alpha)
        maxt = np.quantile(values_values,1-alpha)
    else : 
        mint = min_value
        maxt = max_value
    values = values.clip(mint,maxt)
    print("Extremas of the "+name_values+" plotted : ",mint,maxt)
    values-=np.amin(mint)
    values/=np.amax(maxt-mint)
    
    ps_mesh = view_faces_values_on_embryo(Mesh,values,ps_mesh = ps_mesh,name_values = name_values,colormap = cmap,clean_before = clean_before, clean_after=clean_after,show=show,adapt_values=False,scattered= scattered)

    return(ps_mesh)
        
def plot_trijunctions(Mesh,Dict_trijunctional_values=None,clean_before = True, clean_after = True, show=True, color = "values",value_color = np.ones(3),cmap = cm.jet) : 
    #Dict_trijunctional_values=dict_line_tensions
    Dict_trijunctions = {}

    for edge in Mesh.half_edges : 
        if len(edge.twin)>1 :
            list_materials = []
            for a in edge.twin : 
                list_materials.append(Mesh.half_edges[a].incident_face.material_1)
                list_materials.append(Mesh.half_edges[a].incident_face.material_2)
            list_materials = np.unique(list_materials)
            key_junction = tuple(list_materials)
            Dict_trijunctions[key_junction] = Dict_trijunctions.get(key_junction,[]) + [[edge.origin.key,edge.destination.key]]

    ps.init()
    if clean_before : ps.remove_all_structures()

    if color == "uniform" : 
        plotted_edges = []
        edges = []
        verts = []
        i=0
        for key in Dict_trijunctions:
            if len(key)>=3: 
                for n in range(len(np.array(Dict_trijunctions[key]))): 
                    plotted_edges.append([2*i,2*i+1])
                    i+=1
                edges.append(np.array(Dict_trijunctions[key]))
                verts.append(Mesh.v[edges[-1]].reshape(-1,3))
        
        edges = np.vstack(edges)
        verts = np.vstack(verts)
        plotted_edges = np.array(plotted_edges)
        ps.register_curve_network("trijunctions", verts,plotted_edges,color = value_color)
        
    else : 
        
        edges = []
        verts = []
        plotted_edges=[]
        edges_values = []
        i=0
        for key in Dict_trijunctions : 
            if len(key)>=3:
                edges.append(np.array(Dict_trijunctions[key]))
                verts.append(Mesh.v[edges[-1]].reshape(-1,3))
                for n in range(len(edges[-1])): 
                    plotted_edges.append([2*i,2*i+1])
                    i+=1
                    edges_values.append(Dict_trijunctional_values.get(key,0))

        edges = np.vstack(edges)
        verts = np.vstack(verts)
        plotted_edges = np.array(plotted_edges)
        curv_net = ps.register_curve_network("trijunctions", verts,plotted_edges,color = np.random.rand(3))

        edges_values = np.array(edges_values)
        edges_values/= np.amax(edges_values)
        color_values = cmap(edges_values)[:,:3]
        curv_net.add_color_quantity("line tensions", color_values, defined_on='edges',enabled = True)



    if show : 
        ps.show()
        
    if clean_after : ps.remove_all_structures()






def plot_trijunctions_topo_change_viewer(Mesh,Dict_trijunctional_values=None,clean_before = True, clean_after = True, show=True, color = "values",value_color = np.ones(3)) : 

    Dict_trijunctions = {}

    for edge in Mesh.half_edges : 
        if len(edge.twin)>1 :
            list_materials = []
            for a in edge.twin : 
                list_materials.append(Mesh.half_edges[a].incident_face.material_1)
                list_materials.append(Mesh.half_edges[a].incident_face.material_2)
            list_materials = np.unique(list_materials)
            key_junction = tuple(list_materials)
            Dict_trijunctions[key_junction] = Dict_trijunctions.get(key_junction,[]) + [[edge.origin.key,edge.destination.key]]

    ps.init()
    if clean_before : ps.remove_all_structures()

    if color == "uniform" : 
        plotted_edges = []
        edges = []
        verts = []
        i=0
        for key in Dict_trijunctions:
            if len(key)>=3: 
                for n in range(len(np.array(Dict_trijunctions[key]))): 
                    plotted_edges.append([2*i,2*i+1])
                    i+=1
                edges.append(np.array(Dict_trijunctions[key]))
                verts.append(Mesh.v[edges[-1]].reshape(-1,3))
        
        edges = np.vstack(edges)
        verts = np.vstack(verts)
        plotted_edges = np.array(plotted_edges)
        ps.register_curve_network("trijunctions", verts,plotted_edges,color = value_color)
        
    else : 
        edges_r = []
        verts_r = []
        edges_g = []
        verts_g = []
        green_edges = []
        red_edges = []
        edges_values = []
        b=0
        r=0
        for key in Dict_trijunctions : 
            if len(key)>=3:
                edges_values.append(np.clip(Dict_trijunctional_values.get(key,0),0,None))
                if edges_values[-1]==1: 
                    for n in range(len(Dict_trijunctions[key])):
                        red_edges.append([2*r,2*r+1])
                        r+=1
                    edges_r.append(np.array(Dict_trijunctions[key]))
                    verts_r.append(Mesh.v[edges_r[-1]].reshape(-1,3))
                else : 
                    for n in range(len(Dict_trijunctions[key])):
                        green_edges.append([2*b,2*b+1])
                        b+=1


                    edges_g.append(np.array(Dict_trijunctions[key]))
                    verts_g.append(Mesh.v[edges_g[-1]].reshape(-1,3))


        verts_r = np.vstack(verts_r)
        verts_g = np.vstack(verts_g)
        red_edges = np.array(red_edges)
        green_edges = np.array(green_edges)
        curv_net = ps.register_curve_network("valid trijunctions", verts_g,green_edges,color = (0,1,0),transparency = 0.3)
        curv_net = ps.register_curve_network("bad trijunctions", verts_r,red_edges,color = (1,0,0),transparency = 1.0)


    if show : 
        ps.show()
        
    if clean_after : ps.remove_all_structures()

















import numpy as np 
from scipy import linalg

def change_idx_cells(faces,mapping):
    #mapping : {key_init:key_end} has to be a bijection
    new_faces = faces.copy()
    for key in mapping : 
        new_faces[faces[:,3]==key][:,3]=mapping[key]
        new_faces[faces[:,4]==key][:,4]=mapping[key]
    return(new_faces)


def infer_pressure_laplace(Mesh,dict_tensions,P0=0,weighted=False):
    dict_areas = Mesh.compute_areas_interfaces()
    dict_curvature = Mesh.compute_curvatures_interfaces(weighted=weighted)
    
    M,B = build_matrix_laplace(dict_curvature,dict_areas,dict_tensions,Mesh.n_materials,Mesh.materials,P0)
    x, resid, rank, sigma = linalg.lstsq(M, B)
    
    nc = Mesh.n_materials
    
    dict_pressures = {0:P0}
    for i,key in enumerate(Mesh.materials[1:]): 
        dict_pressures[key]=x[i+1]+P0
        
    return(x,dict_pressures,resid)


def infer_pressure_variational(Mesh,dict_tensions,P0 = 0,torch = True,prints = False):
    DA,DV = Mesh.compute_area_derivatives(),Mesh.compute_volume_derivatives()
    
    M,B,nm= build_matrix_discrete(DA,DV,Mesh.materials)
    
    F_g,F_p,G_g,G_p=extract_submatrices(M,nm)
    
    

    Mat_t = F_g - F_p@(np.linalg.inv(G_p))@G_g
    if prints : 
        print(F_g.shape,F_p.shape,G_g.shape,G_p.shape)
        print("Det of F_g:",np.linalg.det(F_g))
        print("Det of G_p:",np.linalg.det(G_p))

    Mat_inference = np.zeros((nm+1,nm))
    Mat_inference[:nm,:nm]=Mat_t
    Mat_inference[nm:]=1
    B_inf = np.zeros(nm+1)
    B_inf[-1]=nm
    
    dict_pressures = {0:P0}
    
    x = np.zeros(len(dict_tensions.keys()))
    for i,key in enumerate(sorted(list(dict_tensions.keys()))):
        x[i] = dict_tensions[key]
        
    P=np.linalg.inv(G_p)@G_g@x
        
    for i,key in enumerate(Mesh.materials[1:]): 
        dict_pressures[key]=P[i]+P0
    
    return(x,dict_pressures,0)

def infer_pressure(Mesh,dict_tensions,mode='Variational',P0=0,weighted=False):
    if mode =="Laplace":
        return(infer_pressure_laplace(Mesh,dict_tensions,P0,weighted=weighted))
    elif mode =='Variational': 
        return(infer_pressure_variational(Mesh,dict_tensions,P0))
    else : 
        print("Unimplemented method")
        return(None)

def build_matrix_laplace(dict_curvature,dict_areas,dict_tensions,n_materials,materials,P0=0):

    Total_area_mesh = sum(dict_areas.values()) #sum of the areas of all the surfaces of the mesh
    T = np.array(list(dict_curvature.keys()))
    kt = np.amax(T)+1
    Keys_t = T[:,0]*kt + T[:,1]
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))

    #MX = B
    #x : structure : [0,nc[ : pressions
    nm = len(T)
    nc = n_materials-1

    reverse_map_materials = dict(zip(materials, np.arange(len(materials))))
    
    B = np.zeros(nm + 1)
    B[-1]=P0 #Po = Pb with Pb = 1
    M = np.zeros((nm +1, nc + 1))
    
    #CLASSICAL 
    #LAPLACE
    for i,key in enumerate(Keys_t):
        a = key//kt
        b = key%kt
        curvature_k = dict_curvature[(a,b)]
        tension_k = dict_tensions[(a,b)]

        if np.isnan(curvature_k) : 
            continue

        area_k = dict_areas[(a,b)]
        factor = area_k / Total_area_mesh
        idx_tk = reverse_map_t[key]
        
        B[i]=-2*curvature_k*tension_k  * factor # +2Hk*Tk 
        #  -> Sign is negative because we choose the following convention : 
        #If 1 goes inside 2 the interface viewed from 1 is concave, viewed from 2 it is convex. 
        #The curvature is noted negatively : Hk<0. Besides,  P1 > P2 and Tk > 0 thus laplace relation writes : 
        # P1 - P2 = -2*Hk*Tk  <=> P1 - P2 + 2*Hk*Tk = 0
        am = reverse_map_materials[a]
        bm = reverse_map_materials[b]
        M[i,am] = 1  * factor  #Pi
        M[i,bm] = -1 * factor #Pj
        

    M[-1,0]=1 #Enforces Po = Pb (see B[-2])
    #print(B)
    """
    pM = np.linalg.pinv(M)
    x = pM@B"""
    
    return(M,B)

def build_matrix_discrete(DA,DV,materials,mean_tension=1):
    
    T = np.array(list(DA.keys()))
    kt = np.amax(T)+1
    Keys_t = T[:,0] + T[:,1]*kt
    reverse_map_t = dict(zip(Keys_t, np.arange(len(Keys_t))))

    #MX = B
    #x : structure : ([0,nm[ : tensions) ([nm:nm+nc] : pressions) : (ya,yb,yc,yd...p0,p1,p2..pnc)
    nm = len(T)
    nc = len(materials)-1
    size = (nm + nc +1)*(nm +nc +1)
    B = np.zeros(nm + nc +1)
    B[-1]=nm*mean_tension #Sum of tensions = number of membrane <=> mean of tensions = mean_tensions
    M = np.zeros((nm + nc +1, nm +nc))
    #print("n_j",nj,"\nn_m",nm,"\nn_c",nc,"\nNumber of unknowns ",nm+nc+1)

    ###DISCRETE
    ###YOUNG-DUPRÉ
    ###EQUATIONS
    
    for i,tup_l in enumerate(DA.keys()): 
        a_l,b_l = tup_l
        key_l = a_l + b_l*kt
        
        for tup_m in DA.keys(): 
            a_m,b_m = tup_m
            key_m = a_m + b_m*kt
            idx_tm = reverse_map_t[key_m]
            M[i,idx_tm]=np.sum(DA[tup_m]*DA[tup_l])
            
        for n,key_n in enumerate(materials[1:]): 
            M[i,nm+n] = -np.sum(DV[key_n]*DA[tup_l])
        
    off=nm 
    
    ###DISCRETE
    ###LAPLACE
    ###EQUATIONS
    for i,key_i in enumerate(materials[1:]): 

        for tup in DA.keys(): 
            a,b = tup
            key = a + b*kt
            idx_tk = reverse_map_t[key]
            M[off+i,idx_tk] = np.sum(DA[(a,b)]*DV[key_i])
            
        for n,key_n in enumerate(materials[1:]): 
            M[off+i,nm+n] = -np.sum(DV[key_n]*DV[key_i])

    off+=nc
    
    Lines_sum = np.sum(M,axis=1)
    for i in range(len(M)-1): 
        if Lines_sum[i]!=0 : 
            M[i]/=Lines_sum[i]
    
    M[off,:nm]=1 #Enforces mean of tensions = mean_tensions (see B[-1])
    
    return(M,B,nm)

def extract_submatrices(M,nm): 
    F_g = M[:nm,:nm]
    F_p = - M[:nm,nm:]
    G_g = M[nm:-1,:nm]
    G_p = -M[nm:-1,nm:]
    return(F_g,F_p,G_g,G_p)




import numpy as np

def open_pressure_file(filename,ncells): 
    File = open(filename, 'r+')
    File.readline()
    pressure=[]
    for i in range(ncells):
        File.readline()
        pressure.append(float(File.readline()[:-1]))
        
    return(np.array(pressure)/6)

def open_pressure_dict(filename,ncells): 
    File = open(filename, 'r+')
    File.readline()
    pressure=[]
    for i in range(ncells):
        File.readline()
        pressure.append(float(File.readline()[:-1]))
        
    dict_pressure = {0:0}
    for i in range(ncells): 
        dict_pressure[i+1]=pressure[i]
    return(dict_pressure)

def open_tensions_dict(filename,dict_tensions,target_tension=1): 
    #dict_tensions is the dictionnary of infered tensions
    cells_info = np.load(filename,allow_pickle=True).item()
    
    dict_gt ={}
    for key in sorted(dict_tensions.keys()): 
        if key[0]==0 : 
            dict_gt[key] = cells_info['Gamma'][key[1]-1]
        else : 
            dict_gt[key] = cells_info['YIJ'][key[0]-1,key[1]-1]
    m = sum(dict_gt.values())/len(dict_gt)
    for key in dict_gt.keys(): 
        dict_gt[key]/=(m/target_tension)
    return(dict_gt,m/target_tension)

def renormalize_pressures(dict_pressure,factor): 
    #All the tensions are renormalized to have for mean 1. 
    #We need to normalize the pressures by the same coefficient to be consistent. 
    for key in dict_pressure.keys(): 
        dict_pressure[key]/=factor
    return(dict_pressure)

def open_pressure_dict_renormalized(filename, ncells,factor): 
    dict_pressure = open_pressure_dict(filename,ncells)
    return(renormalize_pressures(dict_pressure,factor))

def loss_l2(d1,d2): 
    #compute the loss between all the elements of two dictionnaries assuming both dictionnaries have the same keys
    loss=0
    n=0
    for key in d1.keys(): 
        if not(key in d2.keys()): 
            continue
        n+=1
        loss+=(d1[key]-d2[key])**2
    return(loss/n)

def loss_l1(d1,d2): 
    #compute the loss between all the elements of two dictionnaries assuming both dictionnaries have the same keys
    loss=0
    n=0
    for key in d1.keys(): 
        if not(key in d2.keys()): 
            continue
        n+=1
        loss+=np.abs(d1[key]-d2[key])
        
    return(loss/n)

def loss_ratio_l2(d1,d2): 
    #compute the loss between all the elements of two dictionnaries assuming both dictionnaries have the same keys
    #d2 is the ground truth by convention
    #print("Make sure that you put the ground truth as second argument")
    losses = []
    for key in d1.keys(): 
        if not(key in d2.keys()): 
            continue
        if key == 0 : continue
        losses.append(((d1[key]-d2[key])/d2[key])**2)
    losses = np.array(losses)
    return(np.mean(losses))


def loss_ratio_l1(d1,d2): 
    #compute the loss between all the elements of two dictionnaries assuming both dictionnaries have the same keys
    #d2 is the ground truth by convention
    #print("Make sure that you put the ground truth as second argument")
    losses = []
    for key in d1.keys(): 
        if not(key in d2.keys()): 
            continue
        if key == 0 : continue
        losses.append(np.abs((d1[key]-d2[key])/d1[key]))
    return(np.mean(losses))

def loss_l2_normalized(d1,d2): 
    #compute the loss between all the elements of two dictionnaries assuming both dictionnaries have the same keys
    #d2 is the ground truth by convention
    #print("Make sure that you put the ground truth as second argument")
    losses = []
    mean_d2 = np.mean(list(d2.values()))
    for key in d1.keys(): 
        if key == 0 : continue
        losses.append(((d1[key]-d2[key])/mean_d2)**2)
    losses = np.array(losses)
    return(losses, np.mean(losses))


def loss_l1_normalized(d1,d2): 
    #compute the loss between all the elements of two dictionnaries assuming both dictionnaries have the same keys
    #d2 is the ground truth by convention
    #print("Make sure that you put the ground truth as second argument")
    losses = []
    mean_d2 = np.mean(list(d2.values()))
    for key in d1.keys(): 
        
        if key == 0 : continue
        losses.append(np.abs((d1[key]-d2[key])/mean_d2))
    return(losses, np.mean(losses))




from .Force_viewer import plot_force_inference, plot_tension_inference, plot_residual_junctions, plot_valid_junctions
from .Tension_inference import infer_forces, infer_tension
from .Pressure_inference import infer_pressure